@@ -30,6 +30,16 @@ class TestBasic(unittest.TestCase):
         self.f = np.random.randint(0, 10, size=(20, 3), dtype=self.f1.dtype)
         self.g = np.random.randint(0, 10, size=(20, 4), dtype="int32")
 
+    def tearDown(self):
+        vv1, ff1 = igl.read_triangle_mesh(os.path.join(self.test_path, "bunny_small.off"))
+        vv2, ff2 = igl.read_triangle_mesh(os.path.join(self.test_path, "fertility.off"))
+        self.assertTrue((vv1 == self.v1).all())
+        self.assertTrue((ff1 == self.f1).all())
+
+        self.assertTrue((vv2 == self.v2).all())
+        self.assertTrue((ff2 == self.f2).all())
+
+
     def test_z_module(self):
         # Extract all implemented functions from the module
         funcs = []
@@ -77,6 +87,13 @@ class TestBasic(unittest.TestCase):
         self.assertTrue(l.dtype == self.v.dtype)
         self.assertTrue(type(l) == csc.csc_matrix)
 
+    def test_cotmatrix_intrinsic(self):
+        el = igl.edge_lengths(self.v, self.f)
+        l = igl.cotmatrix_intrinsic(el, self.f)
+        self.assertTrue(l.shape == (self.v.shape[0], self.v.shape[0]))
+        self.assertTrue(l.dtype == el.dtype)
+        self.assertTrue(type(l) == csc.csc_matrix)
+
     def test_ears(self):
         ears, ears_opp = igl.ears(self.f1)
         self.assertTrue(ears.shape == ears_opp.shape)
@@ -93,6 +110,12 @@ class TestBasic(unittest.TestCase):
         self.assertTrue(g.flags.c_contiguous)
 
     # sparse matrix, no flag attribute
+    def test_grad_intrinsic(self):
+        el = igl.edge_lengths(self.v, self.f)
+        g = igl.grad_intrinsic(el, self.f)
+        self.assertTrue(g.shape == (self.f.shape[0] * 2, self.v.shape[0]))
+        self.assertTrue(type(g) == csc.csc_matrix)
+
     def test_grad(self):
         g = igl.grad(self.v, self.f)
         h = igl.grad(self.v, self.f, uniform=True)
@@ -110,6 +133,16 @@ class TestBasic(unittest.TestCase):
         self.assertTrue(a.dtype == np.float64)
         self.assertTrue(type(a) == type(b) == csc.csc_matrix)
 
+    def test_massmatrix_intrinsic(self):
+        el = igl.edge_lengths(self.v, self.f)
+        a = igl.massmatrix_intrinsic(el, self.f)
+        b = igl.massmatrix_intrinsic(el, self.f, type=igl.MASSMATRIX_TYPE_BARYCENTRIC)
+        self.assertTrue(a.shape == (self.v.shape[0], self.v.shape[0]))
+        self.assertTrue(b.shape == (self.v.shape[0], self.v.shape[0]))
+        self.assertTrue(b.dtype == np.float64)
+        self.assertTrue(a.dtype == np.float64)
+        self.assertTrue(type(a) == type(b) == csc.csc_matrix)
+
     def test_principal_curvature(self):
         pd1, pd2, pv1, pv2 = igl.principal_curvature(self.v, self.f)
         qd1, qd2, qv1, qv2 = igl.principal_curvature(self.v, self.f, radius=7, use_k_ring=False)
@@ -209,8 +242,6 @@ class TestBasic(unittest.TestCase):
     def test_adjacency_list(self):
         a = igl.adjacency_list(self.f1)
         self.assertEqual(len(a), self.v1.shape[0])
-        #FIXME: list object has no attribute 'flags'
-        #self.assertTrue(a.flags.c_contiguous)
 
     # all are sparse matrices
     def test_arap_linear_block(self):
@@ -375,14 +406,6 @@ class TestBasic(unittest.TestCase):
         self.assertEqual(p.shape[0], self.g.shape[0])
         self.assertTrue(p.flags.c_contiguous)
 
-    # def test_circulation(self):
-    #    difficult data
-    #    _, e, emap = igl.crouzeix_raviart_cotmatrix(self.v1, self.f1)
-    #    ret = igl.circulation(0, False, emap, e, e)
-    #    self.assertTrue(type(ret) == list)
-    #    self.assertTrue(type(ret[0]) == int)
-
-
     def test_collapse_small_triangles(self):
         ff = igl.collapse_small_triangles(self.v, self.f, 0.5)
         self.assertEqual(ff.shape[1], self.f.shape[1])
@@ -1019,6 +1042,13 @@ class TestBasic(unittest.TestCase):
         self.assertEqual(d.dtype, self.v1.dtype)
         self.assertTrue(d.flags.c_contiguous)
 
+    def test_heat_geodesic(self):
+        vs = np.array([0])
+        d = igl.heat_geodesic(self.v1, self.f1, 1.0, vs)
+        # TODO: Should this not return distances for all sources?
+        self.assertEqual(d.dtype, self.v1.dtype)
+        self.assertTrue(d.flags.c_contiguous)
+
     def test_cut_mesh(self):
        cuts = np.random.randint(0, 2, size=self.f1.shape, dtype=self.f1.dtype)
        vcut, fcut = igl.cut_mesh(self.v1, self.f1, cuts)
@@ -1081,7 +1111,7 @@ class TestBasic(unittest.TestCase):
         max_v = np.max(self.v1, axis=0)
         n = 16
         g = np.mgrid[min_v[0]:max_v[0]:complex(n), min_v[1]:max_v[1]:complex(n), min_v[2]:max_v[2]:complex(n)]
-        p = np.vstack(map(np.ravel, g)).T
+        p = np.vstack(list(map(np.ravel, g))).T
         s, i, c = igl.signed_distance(p, self.v1, self.f1)
 
         self.assertEqual(s.shape[0], p.shape[0])
@@ -1559,6 +1589,49 @@ class TestBasic(unittest.TestCase):
         self.assertTrue(f.dtype == self.f.dtype)
         self.assertTrue(f.shape[1] == 3)
 
+    def test_is_delaunay(self):
+        v = np.array([[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [0, 0], [0, -1]])
+        f = igl.delaunay_triangulation(v)
+        d = igl.is_delaunay(v, f)
+        self.assertTrue(d.flags.c_contiguous)
+        self.assertTrue(d.dtype == np.bool)
+        self.assertTrue(d.shape == f.shape)
+        self.assertTrue(np.all(d))
+
+        d1 = igl.is_delaunay(self.v1, self.f1)
+        self.assertTrue(d1.flags.c_contiguous)
+        self.assertTrue(d1.dtype == np.bool)
+        self.assertTrue(d1.shape == self.f1.shape)
+        self.assertFalse(np.all(d1))
+
+
+        v = np.hstack([v, np.zeros((v.shape[0], 1))])
+        el = igl.edge_lengths(v, f)
+        d2 = igl.is_intrinsic_delaunay(el, f)
+        self.assertTrue(d2.flags.c_contiguous)
+        self.assertTrue(d2.dtype == np.bool)
+        self.assertTrue(d2.shape == f.shape)
+        #self.assertTrue(np.all(d2))  #TODO: Why?
+
+    def test_is_intrinsic_delaunay(self):
+        # Tested above
+        pass
+
+    def test_intrinsic_delaunay_triangulation(self):
+        el = igl.edge_lengths(self.v1, self.f1)
+        l, f = igl.intrinsic_delaunay_triangulation(el, self.f1)
+        # print(l, f)
+        self.assertTrue(f.flags.c_contiguous)
+        self.assertTrue(f.dtype == self.f1.dtype)
+        self.assertTrue(f.shape[1] == 3)
+
+    def test_intrinsic_delaunay_triangulation_edges(self):
+        el = igl.edge_lengths(self.v1, self.f1)
+        l, f, e, u_e, emap, ue2e = igl.intrinsic_delaunay_triangulation_edges(el, self.f1)
+        self.assertTrue(f.flags.c_contiguous)
+        self.assertTrue(f.dtype == self.f.dtype)
+        self.assertTrue(f.shape[1] == 3)
+
     def test_edges_to_path(self):
         e = igl.edges(self.f1)
         e = e[:1, :]
@@ -1572,6 +1645,17 @@ class TestBasic(unittest.TestCase):
         self.assertTrue(j.dtype == e.dtype)
         self.assertTrue(k.dtype == e.dtype)
 
+    def test_path_to_edges(self):
+        e1 = igl.path_to_edges(np.array(range(20)), False)
+        e2 = igl.path_to_edges(np.array(range(20)), True)
+        r2 = np.vstack([np.array(range(20)), np.array(range(1,21))]).T
+        r2[19, 1] = 0
+        self.assertTrue(np.allclose(e2, r2))
+        self.assertTrue(e1.flags.c_contiguous)
+        self.assertTrue(e2.flags.c_contiguous)
+        self.assertTrue(e1.dtype == np.int)
+        self.assertTrue(e2.dtype == np.int)
+
     def test_exterior_edges(self):
         e = igl.exterior_edges(self.f1)
 
@@ -1728,6 +1812,17 @@ class TestBasic(unittest.TestCase):
 
         vol = igl.volume_single(a[0, :], b[0, :], c[0, :], d[0, :])
 
+        l = igl.edge_lengths(v, t)
+        vol = igl.volume_from_edges(l)
+        self.assertTrue(vol.flags.c_contiguous)
+        self.assertTrue(vol.dtype == v.dtype)
+        self.assertTrue(len(vol.shape) == 1)
+        self.assertTrue(vol.shape[0] == t.shape[0])
+
+    def test_volume_from_edges(self):
+        #tested in volume
+        pass
+
     def test_volume_from_vertices(self):
         #tested in volume
         pass
@@ -1946,7 +2041,7 @@ class TestBasic(unittest.TestCase):
     def test_sample_edges(self):
         e = igl.edges(self.f1)
         s = igl.sample_edges(self.v1, e, 4)
-        print(e.shape, s.shape)
+        # print(e.shape, s.shape)
 
         self.assertTrue(s.flags.c_contiguous)
         self.assertTrue(s.dtype == self.v1.dtype)
@@ -1989,6 +2084,273 @@ class TestBasic(unittest.TestCase):
         self.assertTrue(f.shape[0] == self.f.shape[0]-2)
         self.assertTrue(f.shape[1] == 3)
 
+    def test_signed_angle(self):
+        a = np.random.rand(2, 1)
+        b = np.random.rand(2, 1)
+        p = np.random.rand(2, 1)
+
+        angle = igl.signed_angle(a, b, p)
+
+        a = np.random.rand(1, 2)
+        b = np.random.rand(1, 2)
+        p = np.random.rand(1, 2)
+
+        angle = igl.signed_angle(a, b, p)
+
+
+    def test_unique_edge_map(self):
+        E, uE, EMAP, uE2E = igl.unique_edge_map(self.f1)
+
+        self.assertTrue(E.flags.c_contiguous)
+        self.assertTrue(uE.flags.c_contiguous)
+        self.assertTrue(EMAP.flags.c_contiguous)
+
+        self.assertTrue(E.dtype == self.f1.dtype)
+        self.assertTrue(uE.dtype == self.f1.dtype)
+        self.assertTrue(EMAP.dtype == self.f1.dtype)
+
+        self.assertTrue(E.shape == (self.f1.shape[0]*3, 2))
+        self.assertTrue(uE.shape[1] == 2)
+        self.assertTrue(E.shape[0] == self.f1.shape[0]*3)
+
+    def test_connected_components(self):
+        a = igl.adjacency_matrix(self.f)
+        comps, c, k = igl.connected_components(a)
+
+        self.assertTrue(c.flags.c_contiguous)
+        self.assertTrue(k.flags.c_contiguous)
+
+        self.assertTrue(c.dtype == self.f1.dtype)
+        self.assertTrue(k.dtype == self.f1.dtype)
+
+        self.assertTrue(c.shape[0] == a.shape[0])
+
+    def test_pso(self):
+        def banana(x):
+            x1 = x[0]
+            x2 = x[1]
+            return x1**4 - 2*x2*x1**2 + x2**2 + x1**2 - 2*x1 + 5
+
+        lb = np.array([-3.0, -1.0])
+        ub = np.array([2.0, 6.0])
+
+        fopt, xopt = igl.pso(banana, lb, ub, max_iters=10, population=10)
+
+        self.assertTrue(xopt.flags.c_contiguous)
+        self.assertTrue(xopt.dtype == lb.dtype)
+        self.assertTrue(xopt.shape == (2, ))
+
+    def test_random_search(self):
+        def banana(x):
+            x1 = x[0]
+            x2 = x[1]
+            return x1**4 - 2*x2*x1**2 + x2**2 + x1**2 - 2*x1 + 5
+
+        lb = np.array([-3.0, -1.0])
+        ub = np.array([2.0, 6.0])
+
+        fopt, xopt = igl.random_search(banana, lb, ub, iters=10)
+
+        self.assertTrue(xopt.flags.c_contiguous)
+        self.assertTrue(xopt.dtype == lb.dtype)
+        self.assertTrue(xopt.shape == (2, ))
+
+    def test_bijective_composite_harmonic_mapping(self):
+        v, f, _ = igl.read_off(os.path.join(self.test_path, "camelhead.off"))
+        b = igl.boundary_loop(f)
+        thetas = np.linspace(0, 2 * np.pi, len(b))[:, np.newaxis]
+        bc = np.concatenate([np.cos(thetas), np.sin(thetas)], axis=1)
+        v2d = igl.harmonic_weights(v, f, b, bc, 1)[:,:2]
+        ret0, mapping0 = igl.bijective_composite_harmonic_mapping(v2d, f, b, bc)
+        self.assertTrue(ret0)
+        self.assertTrue(mapping0.flags.c_contiguous)
+        self.assertTrue(mapping0.dtype == v2d.dtype)
+        self.assertTrue(mapping0.shape == v2d.shape)
+
+        ret1, mapping1 = igl.bijective_composite_harmonic_mapping_with_steps(v2d, f, b, bc, min_steps=1, max_steps=5, num_inner_iters=2, test_for_flips=True)
+        self.assertTrue(ret1)
+        self.assertTrue(mapping1.flags.c_contiguous)
+        self.assertTrue(mapping1.dtype == v2d.dtype)
+        self.assertTrue(mapping1.shape == v2d.shape)
+
+        ret2, mapping2 = igl.bijective_composite_harmonic_mapping_with_steps(v2d, f, b, bc, min_steps=1, max_steps=5, num_inner_iters=2, test_for_flips=False)
+        self.assertTrue(ret2)
+        self.assertTrue(mapping2.flags.c_contiguous)
+        self.assertTrue(mapping2.dtype == v2d.dtype)
+        self.assertTrue(mapping2.shape == v2d.shape)
+
+        self.assertTrue(np.allclose(mapping0, mapping1))
+
+    def test_bijective_composite_harmonic_mapping_with_steps(self):
+        # Tested above
+        pass
+
+    def test_extract_non_manifold_edge_curves(self):
+        _ = igl.extract_non_manifold_edge_curves(self.f1, [])
+        curves = igl.extract_non_manifold_edge_curves(self.f1, [range(10)])
+        self.assertTrue(len(curves) == 1)
+        self.assertTrue(curves[0][0] == 0)
+
+    def test_intrinsic_delaunay_cotmatrix(self):
+        l, l_int, f_int = igl.intrinsic_delaunay_cotmatrix(self.v1, self.f1)
+        # print(l.shape, l_int.shape, f_int.shape)
+        self.assertTrue(l.shape == (self.v1.shape[0], self.v1.shape[0]))
+        self.assertTrue(l.dtype == self.v1.dtype)
+        self.assertTrue(type(l) == csc.csc_matrix)
+
+    def test_cut_to_disk(self):
+        cuts = igl.cut_to_disk(self.f)
+        self.assertTrue(len(cuts) == 9)
+
+    def test_iterative_closest_point(self):
+        r, t = igl.iterative_closest_point(self.v1, self.f1, self.v1, self.f1, 3, 20)
+        self.assertEqual(r.shape, (3, 3))
+        self.assertEqual(t.shape, (3,))
+        self.assertTrue(r.flags.c_contiguous)
+        self.assertTrue(t.flags.c_contiguous)
+        self.assertTrue(r.dtype == t.dtype == self.v.dtype)
+
+    def test_rigid_alignment(self):
+        n = igl.per_vertex_normals(self.v1, self.f1)
+        r, t = igl.rigid_alignment(self.v1, self.v1+1, n)
+        self.assertTrue(np.allclose(r, np.eye(3)))
+        self.assertTrue(np.allclose(t, np.ones(3)))
+        self.assertEqual(r.shape, (3, 3))
+        self.assertEqual(t.shape, (3,))
+        self.assertTrue(r.flags.c_contiguous)
+        self.assertTrue(t.flags.c_contiguous)
+        self.assertTrue(r.dtype == t.dtype == self.v1.dtype)
+
+    def test_sharp_edges(self):
+        se, e, ue, emap, ue2e, sharp = igl.sharp_edges(self.v1, self.f1, np.pi*0.11)
+        self.assertTrue(se.shape[1] == 2)
+        self.assertTrue(ue.shape[1] == 2)
+        self.assertTrue(emap.shape[0] == self.f1.shape[0]*3)
+        self.assertTrue(se.shape[0] == len(sharp))
+        self.assertTrue(se.flags.c_contiguous)
+        self.assertTrue(e.flags.c_contiguous)
+        self.assertTrue(ue.flags.c_contiguous)
+        self.assertTrue(emap.flags.c_contiguous)
+
+    def test_quad_grid(self):
+        v, q, e = igl.quad_grid(3, 3)
+        self.assertTrue(v.shape == (3*3, 2))
+        self.assertTrue(q.shape == (2*2, 4))
+        self.assertTrue(v.flags.c_contiguous)
+        self.assertTrue(q.flags.c_contiguous)
+        self.assertTrue(v.dtype == np.float)
+        self.assertTrue(q.dtype == np.int)
+        self.assertTrue(e.dtype == np.int)
+
+    def test_sparse_voxel_grid(self):
+        def sphere1(point):
+            return np.sqrt(point[0]**2 + point[1]**2 + point[2]**2) - 1.0
+        point = np.array([1.0, 0.0, 0.0])
+        cs, cv, ci = igl.sparse_voxel_grid(point, sphere1, 1.0, 100)
+        self.assertTrue(cv.flags.c_contiguous)
+        self.assertTrue(cv.dtype == np.float)
+        self.assertTrue(cv.shape == (len(cs), 3))
+        self.assertTrue(ci.flags.c_contiguous)
+        self.assertTrue(ci.dtype == np.int)
+        self.assertTrue(ci.shape[1] == 8)
+        
+    def test_topological_hole_fill(self):
+        f = self.f1
+        b = np.array(range(10))
+        h = [range(10, 20)]
+        ff = igl.topological_hole_fill(f, b, h)
+        self.assertTrue(ff.flags.c_contiguous)
+        self.assertTrue(ff.shape[1] == 3)
+        self.assertTrue(ff.dtype == np.int)
+        self.assertTrue(ff.shape[0] != f.shape[0])
+
+    def test_triangulated_grid(self):
+        v, f = igl.triangulated_grid(10, 10)
+        self.assertTrue(v.shape == (100, 2))
+        self.assertTrue(f.shape == (162, 3))
+        self.assertTrue(f.flags.c_contiguous)
+        self.assertTrue(v.flags.c_contiguous)
+        self.assertTrue(v.dtype == np.float)
+        self.assertTrue(f.dtype == np.int)
+
+    def test_unproject_on_line(self):
+       pos = np.array([10., 10.])
+       eye = np.eye(4)
+       viewport = np.array([0., 0., 100., 100.])
+       p = np.array([15.0, 20.0, 13.0])
+       d = np.array([0.1, 0.2, 1.0]) 
+       t, z = igl.unproject_on_line(pos, eye, viewport, p, d)
+       
+       self.assertTrue(z.flags.c_contiguous)
+       self.assertTrue(z.shape == (3, ))
+       self.assertTrue(z.dtype == np.float)
+
+    def test_unproject_on_plane(self):
+       pos = np.array([10., 10.])
+       eye = np.eye(4)
+       viewport = np.array([0., 0., 100., 100.])
+       p = np.array([1.0, 2.0, 3.0, 2.0])
+       z = igl.unproject_on_plane(pos, eye, viewport, p)
+       
+       self.assertTrue(z.flags.c_contiguous)
+       self.assertTrue(z.shape == (3, ))
+       self.assertTrue(z.dtype == np.float)
+
+    def test_fast_winding_number_for_points(self):       
+        xs = np.linspace(-5.0, 5.0, 10)
+        grid = np.meshgrid(xs, xs, xs, indexing='ij')
+        grid = np.stack(grid).reshape(3, -1, order='F').T
+        n = igl.per_vertex_normals(self.v1, self.f1)
+        a = np.ones((n.shape[0], )) / n.shape[0]
+        
+        wn = igl.fast_winding_number_for_points(self.v1, n, a, grid)
+        self.assertTrue(wn.flags.c_contiguous)
+        self.assertTrue(wn.shape == (grid.shape[0], ))
+        self.assertTrue(wn.dtype == np.float)
+
+    def test_fast_winding_number_for_meshes(self):
+        xs = np.linspace(-5.0, 5.0, 10)
+        grid = np.meshgrid(xs, xs, xs, indexing='ij')
+        grid = np.stack(grid).reshape(3, -1, order='F').T
+        
+        wn = igl.fast_winding_number_for_meshes(self.v1, self.f1, grid)
+        self.assertTrue(wn.flags.c_contiguous)
+        self.assertTrue(wn.shape == (grid.shape[0], ))
+        self.assertTrue(wn.dtype == np.float)
+
+    def test_flip_avoiding_line_search(self):
+        def fun(v):
+           return np.random.rand(1)
+
+        energy, vr = igl.flip_avoiding_line_search(self.f1, self.v1[:, :2], self.v1[:, :2], fun, 10.0)
+        self.assertTrue(vr.shape == (self.v1.shape[0], 2))
+        self.assertTrue(vr.flags.c_contiguous)
+
+    def test_edge_flaps(self):
+        e, emap, ef, ei = igl.edge_flaps(self.f2)
+        self.assertTrue(e.shape[1] == ef.shape[1] == ei.shape[1] == 2)
+        self.assertTrue(e.shape[0] == ef.shape[0] == ei.shape[0])
+        self.assertTrue(emap.shape[0] == self.f2.shape[0] * 3)
+        self.assertTrue(np.min(e) >= 0 and np.max(e) < self.v2.shape[0])
+        self.assertTrue(e.flags.c_contiguous)
+        self.assertTrue(emap.flags.c_contiguous)
+        self.assertTrue(ef.flags.c_contiguous)
+        self.assertTrue(ei.flags.c_contiguous)
+        self.assertTrue(e.dtype == emap.dtype == ef.dtype == ei.dtype == np.int)
+
+    def test_circulation(self):
+        pass
+        #e, emap, ef, ei = igl.edge_flaps(self.f2)
+        #fac = igl.circulation(667, True, emap, ef, ei)
+        #print(fac)
+
+    def test_edge_collapse_is_valid(self):
+        pass
+        #e, emap, ef, ei = igl.edge_flaps(self.f2)
+        #emap = emap.reshape(-1, 3)
+        #val = igl.edge_collapse_is_valid(0, self.f2, e, emap, ef, ei)
+        #print(val)
+
 
 if __name__ == '__main__':
     unittest.main()
